import cv2
import numpy as np

'''
處理畫面更新 畫畫面等
'''


def create_canvas(Canvas_Property=(512, 512, 3)):
    # Default is create a black image
    return np.zeros(Canvas_Property, np.uint8)


class draw(object):
    """
    Canvas : The image where you want to draw the shapes
    color : Color of the shape. for BGR, pass it as a tuple, eg: (255,0,0) for blue. For grayscale, just pass the scalar value.
    thickness : Thickness of the line or circle etc. If -1 is passed for closed figures like circles, it will fill the shape. default thickness = 1
    lineType : Type of line, whether 8-connected, anti-aliased line etc. By default, it is 8-connected. cv2.LINE_AA gives anti-aliased line which looks great for curves.

    Canvas：您要繪製形狀的圖像
    color：形狀的顏色。對於BGR，將其作為元組傳遞，例如：(255,0,0)對於藍色。對於灰度，只需傳遞標量值即可。
    厚度：線或圓等的粗細。如果對封閉的圖形（如圓）傳遞-1，它將填充形狀。默認厚度= 1
    lineType：線的類型，是否為8連接線，抗鋸齒線等。默認情況下，為8連接線。 cv2.LINE_AA給出抗鋸齒的線條，看起來非常適合曲線。
    """

    def __init__(self):
        self.canvas = create_canvas()
        self.clear_key = 'c'
        self.clear_flag = True

    def draw_line(self, Point1=(0, 0), Point2=(511, 511), Color=(255, 0, 0), Size=5):
        # Draw a diagonal blue line with thickness of 5 px
        self.canvas = cv2.line(self.canvas, Point1, Point2, Color, Size)

    def draw_rectangle(self, Point1=(384, 0), Point2=(510, 128), Color=(0, 255, 0), Size=3):
        self.canvas = cv2.rectangle(self.canvas, Point1, Point2, Color, Size)

    def draw_circle(self, Point=(447, 63), Radius=63, Color=(0, 0, 255), Size=-1):
        self.canvas = cv2.circle(self.canvas, Point, Radius, Color, Size)

    def draw_ellipse(self, Center_Point=(256, 256), Shaft_Lenth=(100, 50), Rotate_Angle=0, Start_Angle=0, End_Angle=180,
                     Color=255, Size=-1):
        self.canvas = cv2.ellipse(self.canvas, Center_Point, Shaft_Lenth, Rotate_Angle, Start_Angle, End_Angle, Color,
                                  Size)

    def draw_poly_gon(self, Points=np.array([[10, 5], [20, 30], [70, 20], [50, 10]], np.int32).reshape((-1, 1, 2)),
                      Color=(0, 255, 255)):
        # pts = np.array([[10, 5], [20, 30], [70, 20], [50, 10]], np.int32)
        # pts = pts.reshape((-1, 1, 2))
        self.canvas = cv2.polylines(self.canvas, [Points], True, Color)

    def draw_text(self, Text='OpenCV', Point=(10, 500), Font=cv2.FONT_HERSHEY_SIMPLEX, ForntSize=4,
                  Color=(255, 255, 255), Size=2, Line_Type=cv2.LINE_AA):
        cv2.putText(self.canvas, Text, Point, Font, ForntSize, Color, Size, Line_Type)

    def set_clear_key(self, clear_key):
        self.clear_key = clear_key

    def clear_canvas(self, clear_flag=True):
        self.clear_flag = clear_flag

    def show_canvas(self, canvas_name='Canvas'):
        while True:
            cv2.imshow(canvas_name, self.canvas)

            Key = cv2.waitKey(1) & 0xFF

            if Key == ord('q') or Key == 27:
                break

            elif Key == ord(self.clear_key) and self.clear_flag is True:
                self.canvas = create_canvas()

        cv2.destroyAllWindows()



import cv2

'''
影片處理
'''


def show_camera(video_name='video'):
    Cap = cv2.VideoCapture(0)

    while True:
        ret, frame = Cap.read()
        if ret is True:
            cv2.imshow(video_name, frame)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
        else:
            Cap.release()
            cv2.destroyAllWindows()


def show_video(video, video_name='video'):
    Cap = cv2.VideoCapture(video)

    while True:
        ret, frame = Cap.read()
        if ret is True:
            cv2.imshow(video_name, frame)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
        else:
            Cap.release()
            cv2.destroyAllWindows()


def save_video_camera():
    """
    So we capture a video, process it frame-by-frame and we want to save that video.
    For images, it is very simple, just use cv2.imwrite().
    Here a little more work is required.

    This time we create a VideoWriter object. We should specify the output file name (eg: output.avi).
    Then we should specify the FourCC code (details in next paragraph).
    Then number of frames per second (fps) and frame size should be passed. A
    nd last one is isColor flag.
    If it is True, encoder expect color frame, otherwise it works with grayscale frame.

    FourCC is a 4-byte code used to specify the video codec.
    The list of available codes can be found in fourcc.org.
    It is platform dependent. Following codecs works fine for me.

    In Fedora: DIVX, XVID, MJPG, X264, WMV1, WMV2. (XVID is more preferable. MJPG results in high size video. X264 gives very small size video)
    In Windows: DIVX (More to be tested and added)
    In OSX : (I don’t have access to OSX. Can some one fill this?)
    FourCC code is passed as cv2.VideoWriter_fourcc('M','J','P','G') or cv2.VideoWriter_fourcc(*'MJPG) for MJPG.
    """
    cap = cv2.VideoCapture(0)

    # Define the codec and create VideoWriter object
    fourcc = cv2.VideoWriter_fourcc(*'XVID')
    out = cv2.VideoWriter('output.avi', fourcc, 20.0, (640, 480))

    while cap.isOpened():
        ret, frame = cap.read()
        if ret is True:
            frame = cv2.flip(frame, 0)

            # write the flipped frame
            out.write(frame)
            cv2.imshow('frame', frame)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
        else:
            break

    # Release everything if job is finished
    cap.release()
    out.release()
    cv2.destroyAllWindows()



import cv2

'''
計算及更改執行效率
'''


def get_fps(function):
    time1 = cv2.getTickCount()
    function()
    time2 = cv2.getTickCount()
    return (time2 - time1) / cv2.getTickFrequency()


'''
Many of the OpenCV functions are optimized using SSE2, AVX etc. 
It contains unoptimized code also.
So if our system support these features, we should exploit them (almost all modern day processors support them).
It is enabled by default while compiling. So OpenCV runs the optimized code if it is enabled, else it runs the unoptimized code.
You can use cv2.useOptimized() to check if it is enabled/disabled and cv2.setUseOptimized() to enable/disable it.
Let’s see a simple example.
'''


def set_optimized():
    cv2.setUseOptimized(not cv2.useOptimized())



import cv2

'''
基本圖像處理用
'''


# 取得圖像 行 列 通道數
def get_image_properties(image):
    total = [image.shape, image.size, image.dtype]
    return total


def get_image_shape(image):
    return image.shape


# 取得 圖片大小
def get_image_size(image):
    return image.size


# 取得圖片類型
def get_image_type(image):
    return image.dtype


# 分割通道
def split_image(image):
    B, G, R = cv2.split(image)
    return [B, G, R]


'''
The B,G,R channels of an image can be split into their individual planes when needed. Then,
the individual channels can be merged back together to form a BGR image again. This can be performed by:

b = img[:,:,0]

Suppose, you want to make all the red pixels to zero, you need not split like this and put it equal to zero. 
You can simply use Numpy indexing which is faster.

 img[:,:,2] = 0
'''


# 組合通道
def merge_image(B, G, R):
    return cv2.merge((B, G, R))


# 合併2張圖片 採用透明度
def image_Blending(image1, image1_Alpha, image2, image2_Alpha):
    return cv2.addWeighted(image1, image1_Alpha, image2, image2_Alpha, 0)



import cv2
import numpy as np

'''
物件追蹤
'''


def mean_shift(video, right=100, height=400, c=200, width=125):
    cap = cv2.VideoCapture(video)

    # take first frame of the video
    ret, frame = cap.read()

    # setup initial location of window
    r, h, c, w = right, height, c, width  # simply hardcoded the values
    track_window = (c, r, w, h)

    # set up the ROI for tracking
    roi = frame[r:r + h, c:c + w]
    hsv_roi = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
    mask = cv2.inRange(hsv_roi, np.array((0., 60., 32.)), np.array((180., 255., 255.)))
    roi_hist = cv2.calcHist([hsv_roi], [0], mask, [180], [0, 180])
    cv2.normalize(roi_hist, roi_hist, 0, 255, cv2.NORM_MINMAX)

    # Setup the termination criteria, either 10 iteration or move by atleast 1 pt
    term_crit = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 1)

    while True:
        ret, frame = cap.read()

        if ret is True:
            hsv = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
            dst = cv2.calcBackProject([hsv], [0], roi_hist, [0, 180], 1)

            # apply meanshift to get the new location
            ret, track_window = cv2.meanShift(dst, track_window, term_crit)

            # Draw it on image
            x, y, w, h = track_window
            img2 = cv2.rectangle(frame, (x, y), (x + w, y + h), 255, 2)
            cv2.imshow('img2', img2)

            k = cv2.waitKey(60) & 0xff
            if k == 27:
                break
            else:
                pass

        else:
            break

    cv2.destroyAllWindows()
    cap.release()


def cam_shift(video, right=100, height=200, c=200, width=125):
    cap = cv2.VideoCapture(video)

    # take first frame of the video
    ret, frame = cap.read()

    # setup initial location of window
    r, h, c, w = right, height, c, width  # simply hardcoded the values
    track_window = (c, r, w, h)

    # set up the ROI for tracking
    roi = frame[r:r + h, c:c + w]
    hsv_roi = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
    mask = cv2.inRange(hsv_roi, np.array((0., 60., 32.)), np.array((180., 255., 255.)))
    roi_hist = cv2.calcHist([hsv_roi], [0], mask, [180], [0, 180])
    cv2.normalize(roi_hist, roi_hist, 0, 255, cv2.NORM_MINMAX)

    # Setup the termination criteria, either 10 iteration or move by atleast 1 pt
    term_crit = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 1)

    while True:
        ret, frame = cap.read()

        if ret is True:
            hsv = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
            dst = cv2.calcBackProject([hsv], [0], roi_hist, [0, 180], 1)

            # apply meanshift to get the new location
            ret, track_window = cv2.CamShift(dst, track_window, term_crit)

            # Draw it on image
            pts = cv2.boxPoints(ret)
            pts = np.int0(pts)
            img2 = cv2.polylines(frame, [pts], True, 255, 2)
            cv2.imshow('img2', img2)

            k = cv2.waitKey(60) & 0xff
            if k == 27:
                break
            else:
                pass

        else:
            break

    cv2.destroyAllWindows()
    cap.release()


'''
How to find HSV values to track?
>>> green = np.uint8([[[0,255,0 ]]])
>>> hsv_green = cv2.cvtColor(green,cv2.COLOR_BGR2HSV)
>>> print (hsv_green)
[[[ 60 255 255]]]
'''


def tracking_color(video, track_range_lower=None, track_range_upper=None):
    if track_range_upper is None:
        track_range_upper = [130, 255, 255]
    if track_range_lower is None:
        track_range_lower = [110, 50, 50]
    cap = cv2.VideoCapture(video)

    while True:

        # Take each frame
        _, frame = cap.read()

        # Convert BGR to HSV
        hsv = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)

        # define range of blue color in HSV
        lower_blue = np.array(track_range_lower)
        upper_blue = np.array(track_range_upper)

        # Threshold the HSV image to get only blue colors
        mask = cv2.inRange(hsv, lower_blue, upper_blue)

        # Bitwise-AND mask and original image
        res = cv2.bitwise_and(frame, frame, mask=mask)

        cv2.imshow('frame', frame)
        cv2.imshow('mask', mask)
        cv2.imshow('res', res)
        k = cv2.waitKey(1) & 0xFF
        if k == 27:
            break
        elif k == ord('q'):
            break

    cv2.destroyAllWindows()



import cv2

'''
圖片閥值化
'''

'''
Here, the matter is straight forward. 
If pixel value is greater than a threshold value, it is assigned one value (may be white),
else it is assigned another value (may be black). 
The function used is cv2.threshold. 
First argument is the source image, which should be a grayscale image. 
Second argument is the threshold value which is used to classify the pixel values. 
Third argument is the maxVal which represents the value to be given 
if pixel value is more than (sometimes less than) the threshold value. 
OpenCV provides different styles of thresholding and it is decided by the fourth parameter of the function.
Different types are:

cv2.THRESH_BINARY
cv2.THRESH_BINARY_INV
cv2.THRESH_TRUNC
cv2.THRESH_TOZERO
cv2.THRESH_TOZERO_INV
'''


def basic_thresholding(image, mode, start=127, end=255):
    return cv2.threshold(image, start, end, mode)


'''
Adaptive Method - It decides how thresholding value is calculated.

cv2.ADAPTIVE_THRESH_MEAN_C : threshold value is the mean of neighbourhood area.
 cv2.ADAPTIVE_THRESH_GAUSSIAN_C : threshold value is the weighted sum of neighbourhood values where weights are a gaussian window.

Block Size - It decides the size of neighbourhood area.
 C - It is just a constant which is subtracted from the mean or weighted mean calculated.
  Below piece of code compares global thresholding and adaptive thresholding for an image with varying illumination:
'''


def adaptive_thresholding(image, adaptive_mode, mode, block_size, C, end=255):
    return cv2.adaptiveThreshold(image, end, adaptive_mode, mode, block_size, C)


'''
For this, our cv2.threshold() function is used, but pass an extra flag, cv2.THRESH_OTSU.
 For threshold value, simply pass zero.
  Then the algorithm finds the optimal threshold value and returns you as the second output, retVal.
   If Otsu thresholding is not used, retVal is same as the threshold value you used.
'''


# I applied Otsu’s thresholding directly.
def otsu_thresholding(image, start=0, end=255):
    return cv2.threshold(image, start, end, cv2.THRESH_BINARY + cv2.THRESH_OTSU)


# I filtered image with a 5x5 gaussian kernel to remove the noise, then applied Otsu thresholding.
def otsu_thresholding_gaussian(image, gaussian=(5, 5), start=0, end=255):
    blur = cv2.GaussianBlur(image, gaussian, 0)
    return cv2.threshold(blur, start, end, cv2.THRESH_BINARY + cv2.THRESH_OTSU)



import cv2

'''
依照特徵點比對
'''

'''
For BF matcher, first we have to create the BFMatcher object using cv2.BFMatcher(). 
It takes two optional params. First one is normType. 
It specifies the distance measurement to be used. 
By default, it is cv2.NORM_L2. It is good for SIFT, SURF etc (cv2.NORM_L1 is also there). 
For binary string based descriptors like ORB, BRIEF, BRISK etc, cv2.NORM_HAMMING should be used, 
which used Hamming distance as measurement. If ORB is using VTA_K == 3 or 4, cv2.NORM_HAMMING2 should be used.

Second param is boolean variable, crossCheck which is false by default. 
If it is true, Matcher returns only those matches with value (i,j) 
such that i-th descriptor in set A has j-th descriptor in set B as the best match and vice-versa. 
That is, the two features in both sets should match each other. 
It provides consistent result, and is a good alternative to ratio test proposed by D.Lowe in SIFT paper.
'''

'''
DMatch.distance - Distance between descriptors. The lower, the better it is.
DMatch.trainIdx - Index of the descriptor in train descriptors
DMatch.queryIdx - Index of the descriptor in query descriptors
DMatch.imgIdx - Index of the train image.
'''


# 比對圖片相似度
def similarity(Image_1, Image_2):
    try:
        # 讀圖
        img1 = cv2.imread(Image_1, 0)
        img2 = cv2.imread(Image_2, 0)

        # ORB檢測器
        orb = cv2.ORB_create()
        kp1, des1 = orb.detectAndCompute(img1, None)
        kp2, des2 = orb.detectAndCompute(img2, None)

        # 特徵點
        bf = cv2.BFMatcher(cv2.NORM_HAMMING)

        # knn塞選
        matches = bf.knnMatch(des1, trainDescriptors=des2, k=2)

        '''
        匹配點數目
        比值測試消除異常點，小於閾值時（0.75）才被認為是匹配，
        因為假設匹配是一一對應的，真正的匹配的理想距離為0
        '''
        good = [m for (m, n) in matches if m.distance < 0.75 * n.distance]
        print(len(good))
        print(len(matches))
        similarly = len(good) / len(matches)
        print("相似度:%s" % similarly)
        return similarly

    except Exception as error:
        print(error)



import cv2
import numpy as np

'''
平滑圖像

對圖像進行濾波 會變模糊
'''

'''
As for one-dimensional signals, images also can be filtered with various low-pass filters (LPF), high-pass filters (HPF), etc. 
A LPF helps in removing noise, or blurring the image. A HPF filters helps in finding edges in an image.

OpenCV provides a function, cv2.filter2D(), to convolve a kernel with an image.
As an example, we will try an averaging filter on an image.

kernel = 5x5

Filtering with the above kernel results in the following being performed: 
for each pixel, a 5x5 window is centered on this pixel, all pixels falling within this window are summed up, 
and the result is then divided by 25. 
This equates to computing the average of the pixel values inside that window. 
This operation is performed for all the pixels in the image to produce the output filtered image.

2D Convolution ( image Filtering )
'''


def convolution2_d(image, kernel=(5, 5)):
    return cv2.filter2D(image, -1, np.ones(kernel, np.float32) / 25)


'''
image blurring is achieved by convolving the image with a low-pass filter kernel. 
It is useful for removing noise.
It actually removes high frequency content (e.g: noise, edges)
from the image resulting in edges being blurred when this is filter is applied.
(Well, there are blurring techniques which do not blur edges). OpenCV provides mainly four types of blurring techniques.
  
1. Averaging
This is done by convolving the image with a normalized box filter. 
It simply takes the average of all the pixels under kernel area and replaces the central element with this average. 
This is done by the function cv2.blur() or cv2.boxFilter(). 
Check the docs for more details about the kernel. 
We should specify the width and height of kernel.
'''


def averaging(image, kernel=(5, 5)):
    return cv2.blur(image, kernel)


'''
In this approach, instead of a box filter consisting of equal filter coefficients, a Gaussian kernel is used.
It is done with the function, cv2.GaussianBlur().
We should specify the width and height of the kernel which should be positive and odd.
We also should specify the standard deviation in the X and Y directions, sigmaX and sigmaY respectively.
If only sigmaX is specified, sigmaY is taken as equal to sigmaX. 
If both are given as zeros, they are calculated from the kernel size. Gaussian filtering is highly effective in removing Gaussian noise from the image.
'''


def gaussian_blur(image, kernel=(5, 5)):
    return cv2.GaussianBlur(image, kernel, 0)


'''
Here, the function cv2.medianBlur() computes the median of all the pixels 
under the kernel window and the central pixel is replaced with this median value. 
This is highly effective in removing salt-and-pepper noise. 
One interesting thing to note is that, in the Gaussian and box filters, 
the filtered value for the central element can be a value which may not exist in the original image.
However this is not the case in median filtering, since the central element is always replaced by some pixel value in the image.
This reduces the noise effectively. The kernel size must be a positive odd integer.
'''


def median_blur(image, kernel=5):
    return cv2.medianBlur(image, kernel)


'''
This is not the case for the bilateral filter, cv2.bilateralFilter(),
which was defined for, and is highly effective at noise removal while preserving edges. 
But the operation is slower compared to other filters. 
We already saw that a Gaussian filter takes the a neighborhood around the pixel and finds its Gaussian weighted average. 
This Gaussian filter is a function of space alone, that is, nearby pixels are considered while filtering. 
It does not consider whether pixels have almost the same intensity value and does not consider 
whether the pixel lies on an edge or not. 
The resulting effect is that Gaussian filters tend to blur edges, which is undesirable.

The bilateral filter also uses a Gaussian filter in the space domain, 
but it also uses one more (multiplicative) Gaussian filter component which is a function of pixel intensity differences. 
The Gaussian function of space makes sure that only pixels are ‘spatial neighbors’ are considered for filtering, 
while the Gaussian component applied in the intensity domain (a Gaussian function of intensity differences) ensures 
that only those pixels with intensities similar to that of the central pixel (‘intensity neighbors’) are included 
to compute the blurred intensity value. As a result, this method preserves edges, since for pixels lying near edges, 
neighboring pixels placed on the other side of the edge,
and therefore exhibiting large intensity variations when compared to the central pixel, will not be included for blurring.
'''


def bilateral(image):
    return cv2.bilateralFilter(image, 9, 75, 75)



import cv2
import numpy as np

'''
高斯金字塔

高斯圖像融合

圖片金字塔
'''


# 降階金字塔 圖片會變小
def down_image(image):
    return cv2.pyrDown(image)


# 升階金字塔 圖片會變大
def up_image(image):
    return cv2.pyrUp(image)


'''
One application of Pyramids is Image Blending. 
For example, in image stitching, you will need to stack two images together, 
but it may not look good due to discontinuities between images. 
In that case, image blending with Pyramids gives you seamless blending without leaving much data in the images.

Load the two images of apple and orange
Find the Gaussian Pyramids for apple and orange (in this particular example, number of levels is 6)
From Gaussian Pyramids, find their Laplacian Pyramids
Now join the left half of apple and right half of orange in each levels of Laplacian Pyramids
Finally from this joint image pyramids, reconstruct the original image.

將2張圖片融合
'''


def Blending_Image_Pyramids(image1, image2):
    # generate Gaussian pyramid for A
    global cols
    g = image1.copy()
    gpA = [g]
    for i in range(6):
        g = cv2.pyrDown(g)
        gpA.append(g)

    # generate Gaussian pyramid for B
    g = image2.copy()
    gpB = [g]
    for i in range(6):
        g = cv2.pyrDown(g)
        gpB.append(g)

    # generate Laplacian Pyramid for A
    lpA = [gpA[5]]
    for i in range(5, 0, -1):
        GE = cv2.pyrUp(gpA[i])
        L = cv2.subtract(gpA[i - 1], GE)
        lpA.append(L)

    # generate Laplacian Pyramid for B
    lpB = [gpB[5]]
    for i in range(5, 0, -1):
        GE = cv2.pyrUp(gpB[i])
        L = cv2.subtract(gpB[i - 1], GE)
        lpB.append(L)

    # Now add left and right halves of images in each level
    LS = []
    for la, lb in zip(lpA, lpB):
        rows, cols, dpt = la.shape
        ls = np.hstack((la[:, 0:cols / 2], lb[:, cols / 2:]))
        LS.append(ls)

    # now reconstruct
    ls_ = LS[0]
    for i in range(1, 6):
        ls_ = cv2.pyrUp(ls_)
        ls_ = cv2.add(ls_, LS[i])

    # image with direct connecting each half
    return np.hstack((image1[:, :cols / 2], image2[:, cols / 2:]))



import cv2
import numpy as np

'''
What are contours?
Contours can be explained simply as a curve joining all the continuous points (along the boundary), having same color or intensity. 
The contours are a useful tool for shape analysis and object detection and recognition.

For better accuracy, use binary images. So before finding contours, apply threshold or canny edge detection.
findContours function modifies the source image. 
So if you want source image even after finding contours, already store it to some other variables.
In OpenCV, finding contours is like finding white object from black background. 
So remember, object to be found should be white and background should be black.

得到特徵 並畫出輪廓區域
'''

'''
See, there are three arguments in cv2.findContours() function, 
first one is source image, 
second is contour retrieval mode, 
third is contour approximation method. 
And it outputs the image, contours and hierarchy. contours is a Python list of all the contours in the image. 
Each individual contour is a Numpy array of (x,y) coordinates of boundary points of the object.
'''


def contours_binary_image_all(image, Mode=cv2.RETR_TREE, start=127, end=255):
    image_array = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    ret, thresh = cv2.threshold(image_array, start, end, 0)
    image, contours, hierarchy = cv2.findContours(thresh, Mode, cv2.CHAIN_APPROX_SIMPLE)
    return cv2.drawContours(image, contours, -1, (0, 255, 0), 3)


'''
This is the third argument in cv2.findContours function. What does it denote actually?

Above, we told that contours are the boundaries of a shape with same intensity. 
It stores the (x,y) coordinates of the boundary of a shape. But does it store all the coordinates ? 
That is specified by this contour approximation method.

If you pass cv2.CHAIN_APPROX_NONE, all the boundary points are stored. But actually do we need all the points? 
For eg, you found the contour of a straight line. 
Do you need all the points on the line to represent that line? 
No, we need just two end points of that line. 
This is what cv2.CHAIN_APPROX_SIMPLE does. 
It removes all redundant points and compresses the contour, thereby saving memory.
'''


def contours_binary_image(image, Mode=cv2.RETR_TREE, start=127, end=255):
    image_array = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    ret, thresh = cv2.threshold(image_array, start, end, 0)
    image, contours, hierarchy = cv2.findContours(thresh, Mode, cv2.CHAIN_APPROX_SIMPLE)
    cnt = contours[4]
    return cv2.drawContours(image, [cnt], 0, (0, 255, 0), 3)


'''
1. Moments
image moments help you to calculate some features like center of mass of the object, area of the object etc.

The function cv2.moments() gives a dictionary of all moment values calculated.

From this moments, you can extract useful data like area, centroid etc

M=cv2.moments(cnt)

cx = int(M['m10']/M['m00'])
cy = int(M['m01']/M['m00'])

'''


def moments(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    return cv2.moments(cnt)


'''
2. Contour Area
Contour area is given by the function cv2.contourArea() or from moments, M[‘m00’].
'''


def area(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    return cv2.contourArea(cnt)


'''
3. Contour Perimeter
It is also called arc length. It can be found out using cv2.arcLength() function. 
Second argument specify whether shape is a closed contour (if passed True), or just a curve.
'''


def perimeter(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    return cv2.arcLength(cnt, True)


'''
4. Contour Approximation
It approximates a contour shape to another shape with less number of vertices depending upon the precision we specify. 
It is an implementation of Douglas-Peucker algorithm. 
Check the wikipedia page for algorithm and demonstration.
'''


def approximation(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    epsilon = 0.1 * cv2.arcLength(cnt, True)
    return cv2.approxPolyDP(cnt, epsilon, True)


'''
5. Convex Hull
Convex Hull will look similar to contour approximation, but it is not (Both may provide same results in some cases). 
Here, cv2.convexHull() function checks a curve for convexity defects and corrects it. 
Generally speaking, convex curves are the curves which are always bulged out, or at-least flat. 
And if it is bulged inside, it is called convexity defects. 

hull = cv2.convexHull(points[, hull[, clockwise[, returnPoints]]

Arguments details:
points are the contours we pass into.
hull is the output, normally we avoid it.
clockwise : Orientation flag. If it is True, the output convex hull is oriented clockwise. 
Otherwise, it is oriented counter-clockwise.
returnPoints : By default, True. Then it returns the coordinates of the hull points. 
If False, it returns the indices of contour points corresponding to the hull points.

So to get a convex hull as in above image, following is sufficient:
'''


def hull(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    return cv2.convexHull(cnt)


'''
6. Checking Convexity
There is a function to check if a curve is convex or not, cv2.isContourConvex(). 
It just return whether True or False. Not a big deal.
'''


def convexity(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    return cv2.isContourConvex(cnt)


'''
7. Bounding Rectangle
There are two types of bounding rectangles.

7.a. Straight Bounding Rectangle
It is a straight rectangle, it doesn’t consider the rotation of the object.
So area of the bounding rectangle won’t be minimum. 
It is found by the function cv2.boundingRect().

Let (x,y) be the top-left coordinate of the rectangle and (w,h) be its width and height.
'''


def straight_bounding(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    x, y, w, h = cv2.boundingRect(cnt)
    return cv2.rectangle(image, (x, y), (x + w, y + h), (0, 255, 0), 2)


'''
7.b. Rotated Rectangle
Here, bounding rectangle is drawn with minimum area, so it considers the rotation also. 
The function used is cv2.minAreaRect(). 
It returns a Box2D structure which contains following detals - ( top-left corner(x,y), (width, height), angle of rotation ). 
But to draw this rectangle, we need 4 corners of the rectangle. 
It is obtained by the function cv2.boxPoints()
'''


def rotated(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    rect = cv2.minAreaRect(cnt)
    box = cv2.boxPoints(rect)
    box = np.int0(box)
    return cv2.drawContours(image, [box], 0, (0, 0, 255), 2)


'''
8. Minimum Enclosing Circle
Next we find the circumcircle of an object using the function cv2.minEnclosingCircle(). 
It is a circle which completely covers the object with minimum area.
'''


def minimum_enclosing(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    (x, y), radius = cv2.minEnclosingCircle(cnt)
    center = (int(x), int(y))
    radius = int(radius)
    return cv2.circle(image, center, radius, (0, 255, 0), 2)


'''
9. Fitting a Line
Similarly we can fit a line to a set of points. 
Below image contains a set of white points. 
We can approximate a straight line to it.
'''


def fitting_an_line(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    rows, cols = image.shape[:2]
    [vx, vy, x, y] = cv2.fitLine(cnt, cv2.DIST_L2, 0, 0.01, 0.01)
    lefty = int((-x * vy / vx) + y)
    righty = int(((cols - x) * vy / vx) + y)
    return cv2.line(image, (cols - 1, righty), (0, lefty), (0, 255, 0), 2)


'''
1. Aspect Ratio
It is the ratio of width to height of bounding rect of the object.
長寬比
'''


def aspect_ratio(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    x, y, w, h = cv2.boundingRect(cnt)
    return float(w) / h


'''
2. Extent
Extent is the ratio of contour area to bounding rectangle area.
範圍
'''


def extent(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    extent_area = cv2.contourArea(cnt)
    x, y, w, h = cv2.boundingRect(cnt)
    rect_area = w * h
    return float(extent_area) / rect_area


'''
3. Solidity
Solidity is the ratio of contour area to its convex hull area.
堅固度是輪廓面積與其凸包面積的比率。
'''


def solidity(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    solidity_area = cv2.contourArea(cnt)
    solidity_hull = cv2.convexHull(cnt)
    hull_area = cv2.contourArea(solidity_hull)
    return float(solidity_area) / hull_area


'''
4. Equivalent Diameter
Equivalent Diameter is the diameter of the circle whose area is same as the contour area.
等效直徑
'''


def equivalent_diameter(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    equivalent_diameter_area = cv2.contourArea(cnt)
    return np.sqrt(4 * equivalent_diameter_area / np.pi)


'''
5. Orientation
Orientation is the angle at which object is directed. 
Following method also gives the Major Axis and Minor Axis lengths.
'''


def orientation(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    return cv2.fitEllipse(cnt)


'''
6. Mask and Pixel Points
In some cases, we may need all the points which comprises that object
'''


def mask_and_pixel_points(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    mask = np.zeros(thresh.shape, np.uint8)
    cv2.drawContours(mask, [cnt], 0, 255, -1)
    return np.transpose(np.nonzero(mask))
    # pixelpoints = cv2.findNonZero(mask)


'''
Here, two methods, one using Numpy functions, next one using OpenCV function (last commented line) are given to do the same.
Results are also same, but with a slight difference. 
Numpy gives coordinates in (row, column) format, while OpenCV gives coordinates in (x,y) format. 
So basically the answers will be interchanged. Note that, row = x and column = y.
在這裡，給出了兩種方法，一種使用Numpy函數，另一種使用OpenCV函數（最後註釋的行）執行相同的操作。
結果也相同，但略有不同。Numpy以（行，列）格式給出坐標，而OpenCV以（x，y）格式給出坐標。
因此，基本上答案是可以互換的。
! 注意，row = x，column = y
'''

'''
7. Maximum Value, Minimum Value and their locations
We can find these parameters using a mask image.
'''


def value_locations(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    mask = np.zeros(thresh.shape, np.uint8)
    return cv2.minMaxLoc(thresh, mask=mask)


'''
8. Mean Color or Mean Intensity
Here, we can find the average color of an object. 
Or it can be average intensity of the object in grayscale mode. 
We again use the same mask to do it.
'''


def intensity(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    mask = np.zeros(thresh.shape, np.uint8)
    return cv2.mean(image, mask=mask)


'''
9. Extreme Points
Extreme Points means topmost, bottommost, rightmost and leftmost points of the object.
'''


def extreme(image, start=127, end=255):
    ret, thresh = cv2.threshold(image, start, end, 0)
    contours, hierarchy = cv2.findContours(thresh, 1, 2)
    cnt = contours[0]
    leftmost = tuple(cnt[cnt[:, :, 0].argmin()][0])
    rightmost = tuple(cnt[cnt[:, :, 0].argmax()][0])
    topmost = tuple(cnt[cnt[:, :, 1].argmin()][0])
    bottommost = tuple(cnt[cnt[:, :, 1].argmax()][0])
    return leftmost, rightmost, topmost, bottommost


'''
It returns an array where each row contains these values
[ start point, end point, farthest point, approximate distance to farthest point ].
We can visualize it using an image. We draw a line joining start point and end point, then draw a circle at the farthest point. 
Remember first three values returned are indices of cnt. So we have to bring those values from cnt.
'''


def convexity_defects(image):
    img_gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    ret, thresh = cv2.threshold(img_gray, 127, 255, 0)
    contours, hierarchy = cv2.findContours(thresh, 2, 1)
    cnt = contours[0]
    convexity_defects_hull = cv2.convexHull(cnt, returnPoints=False)
    return cv2.convexityDefects(cnt, convexity_defects_hull)


'''
OpenCV comes with a function cv2.matchShapes() which enables us to compare two shapes, 
or two contours and returns a metric showing the similarity. The lower the result, the better match it is. 
It is calculated based on the hu-moment values. Different measurement methods are explained in the docs.
'''


def match_shapes(image1, image2):
    ret, thresh = cv2.threshold(image1, 127, 255, 0)
    ret, thresh2 = cv2.threshold(image2, 127, 255, 0)
    contours, hierarchy = cv2.findContours(thresh, 2, 1)
    cnt1 = contours[0]
    contours, hierarchy = cv2.findContours(thresh2, 2, 1)
    cnt2 = contours[0]
    return cv2.matchShapes(cnt1, cnt2, 1, 0.0)



import cv2

'''
OpenCV comes with a trainer as well as detector. 
If you want to train your own classifier for any object like car, planes etc. 
you can use OpenCV to create one. 
Its full details are given here: Cascade Classifier Training.

OpenCV already contains many pre-trained classifiers for face, eyes, smile etc. 
Those XML files are stored in opencv/data/haarcascades/ folder. 
Let’s create face and eye detector with OpenCV.
'''


def detection(Image):
    face_cascade = cv2.CascadeClassifier('haarcascade_frontalface_default.xml')
    eye_cascade = cv2.CascadeClassifier('haarcascade_eye.xml')
    gray = cv2.cvtColor(Image, cv2.COLOR_BGR2GRAY)
    faces = face_cascade.detectMultiScale(gray, 1.3, 5)
    for (x, y, w, h) in faces:
        Image = cv2.rectangle(Image, (x, y), (x + w, y + h), (255, 0, 0), 2)
        roi_gray = gray[y:y + h, x:x + w]
        roi_color = Image[y:y + h, x:x + w]
        eyes = eye_cascade.detectMultiScale(roi_gray)
        for (ex, ey, ew, eh) in eyes:
            cv2.rectangle(roi_color, (ex, ey), (ex + ew, ey + eh), (0, 255, 0), 2)
    return Image



import cv2

'''
基本的輸出入圖片
'''


def read_image(image, flag=1):
    return cv2.imread(image, flag)


def show_image(image, window_name='image'):
    cv2.imshow(window_name, image)
    cv2.waitKey()
    cv2.destroyAllWindows()


def output_image(image, file_name='image'):
    cv2.imwrite(file_name, image)



import cv2
import numpy as np


def get_hist_open_cv(Image):
    return cv2.calcHist([Image], [0], None, [256], [0, 256])


def get_hist_num_py(Image):
    return np.histogram(Image.ravel(), 256, [0, 256])



import cv2
import numpy as np

'''
Numpy has an FFT package to do this. 
np.fft.fft2() provides us the frequency transform which will be a complex array. 
Its first argument is the input image, which is grayscale. 
Second argument is optional which decides the size of output array. 
If it is greater than size of input image, input image is padded with zeros before calculation of FFT. 
If it is less than input image, input image will be cropped. 
If no arguments passed, Output array size will be same as input.

Now once you got the result, zero frequency component (DC component) will be at top left corner. 
If you want to bring it to center, you need to shift the result by \frac{N}{2} in both the directions. 
This is simply done by the function, np.fft.fftshift(). 
(It is more easier to analyze). 
Once you found the frequency transform, you can find the magnitude spectrum.
'''


def fourier_numpy(image):
    f = np.fft.fft2(image)
    fshift = np.fft.fftshift(f)
    return 20 * np.log(np.abs(fshift))


'''
OpenCV provides the functions cv2.dft() and cv2.idft() for this. 
It returns the same result as previous, but with two channels. 
First channel will have the real part of the result and second channel will have the imaginary part of the result. 
The input image should be converted to np.float32 first.
'''


def fourier_open_cv(image):
    dft = cv2.dft(np.float32(image), flags=cv2.DFT_COMPLEX_OUTPUT)
    dft_shift = np.fft.fftshift(dft)
    return 20 * np.log(cv2.magnitude(dft_shift[:, :, 0], dft_shift[:, :, 1]))



import cv2
import numpy as np

'''
對圖片的物件進行尋找
'''

'''
  原理:
  輸入兩張影像，分別為 image、template
  不斷滑動 template，得到 image 上各個位置的比較值，比較值代表相似程度
  然後將 image 左上角位置，作為 result 比較值的存放位置
  完成後可得到 result
  可用 minMaxLoc() 函式，找出結果圖的最大或最小值，定位出搜尋位置

  限制 :

  物體有旋轉時，會找不到
  物體大小改變時，會找不到

  參數
  image-被尋找的圖片-必須為 8-bit or 32-bit

  template-尋找的物品圖片

  size 不能大於 image，且格式需一致

  method-比對的方法

  result-比較的結果，格式為 numpy.ndarray (dtype=float32)-可傳入想儲存結果的 array

  CV_TM_SQDIFF : 平方差，越小越相似

  CV_TM_SQDIFF_NORMED : 正規化平方差，越小越相似 保證當 pixel 亮度都乘上同一係數時，相似度不變

  CV_TM_CCORR : 相關係數，越大越相似

  CV_TM_CCORR_NORMED : 正規化相關係數，越大越相似 保證當 pixel 亮度都乘上同一係數時，相似度不變

  CV_TM_CCOEFF : 去掉直流成份的相關係數，越大越相似

  CV_TM_CCOEFF_NORMED : 正規化 去掉直流成份的相關係數 保證當 pixel 亮度都乘上同一係數時，相似度不變
  計算出的相關係數被限制在了 -1 到 1 之間
  1 表示完全相同
  -1 表示亮度正好相反
  0 表示没有線性相關

  詳情 :
  https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_imgproc/py_template_matching/py_template_matching.html?highlight=matchtemplate
  '''


def detect(image, template):
    global bottom_right, top_left
    temp = (0, 0)
    count = 0
    image2 = image.copy()
    w = template.shape[1]
    h = template.shape[0]
    flag = False

    # All the 6 methods for comparison in a list
    methods = ['cv2.TM_CCOEFF', 'cv2.TM_CCOEFF_NORMED', 'cv2.TM_CCORR',
               'cv2.TM_CCORR_NORMED', 'cv2.TM_SQDIFF', 'cv2.TM_SQDIFF_NORMED']

    for meth in methods:
        image = image2.copy()
        method = eval(meth)

        # Apply template Matching
        # res = cv2.matchTemplate(image, template, method) some time this will find error image
        res = cv2.matchTemplate(image, template, TM_CCOEFF_NORMED)
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
        # If the method is TM_SQDIFF or TM_SQDIFF_NORMED, take minimum
        if method in [cv2.TM_SQDIFF, cv2.TM_SQDIFF_NORMED]:
            top_left = min_loc
        else:
            top_left = max_loc
        bottom_right = (top_left[0] + w, top_left[1] + h)

        if count == 0:
            temp = top_left
            count += 1
        else:
            if temp == top_left:
                count += 1
        cv2.rectangle(image, top_left, bottom_right, 255, 2)
        flag = True
    return [image, flag, top_left, bottom_right]


def detect_multi(image, template):
    w, h = template.shape[::-1]
    flag = False
    res = cv2.matchTemplate(image, template, cv2.TM_CCOEFF_NORMED)
    threshold = 0.8
    loc = np.where(res >= threshold)
    for pt in zip(*loc[::-1]):
        cv2.rectangle(image, pt, (pt[0] + w, pt[1] + h), (0, 0, 255), 2)
        flag = True
    return [image, flag]


# 尋找圖中的物件
def find_object_cv2(image, template):
    image = cv2.imread(image, 0)
    template = cv2.imread(template, 0)
    return detect(image, template)


def find_object_cv2_with_pil(image, template):
    image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2GRAY)
    template = cv2.imread(template, 0)
    return detect(image, template)


'''
尋找圖中的多個重複物件
which occurs only once in the image. 
Suppose you are searching for an object which has multiple occurances, cv2.
minMaxLoc() won’t give you all the locations. In that case, we will use thresholding.
'''


def find_multi_object_cv2(image, template):
    image = cv2.imread(image)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    template = cv2.imread(template, 0)
    return detect_multi(image, template)


def find_multi_object_cv2_with_pil(image, template):
    image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2GRAY)
    template = cv2.imread(template, 0)
    return detect_multi(image, template)



import cv2
import numpy as np


def detect(image):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    edges = cv2.Canny(gray, 50, 150, apertureSize=3)

    lines = cv2.HoughLines(edges, 1, np.pi / 180, 200)
    for rho, theta in lines[0]:
        a = np.cos(theta)
        b = np.sin(theta)
        x0 = a * rho
        y0 = b * rho
        x1 = int(x0 + 1000 * (-b))
        y1 = int(y0 + 1000 * a)
        x2 = int(x0 - 1000 * (-b))
        y2 = int(y0 - 1000 * a)

        cv2.line(image, (x1, y1), (x2, y2), (0, 0, 255), 2)
    return image



import cv2
import numpy as np

'''
特徵點檢測
'''

'''
As usual, we have to create an ORB object with the function, cv2.ORB() or using feature2d common interface. 
It has a number of optional parameters. 
Most useful ones are nFeatures which denotes maximum number of features to be retained (by default 500), 
scoreType which denotes whether Harris score or FAST score to rank the features (by default, Harris score) etc. 
Another parameter, WTA_K decides number of points that produce each element of the oriented BRIEF descriptor. 
By default it is two, ie selects two points at a time. 
In that case, for matching, NORM_HAMMING distance is used. 
If WTA_K is 3 or 4, which takes 3 or 4 points to produce BRIEF descriptor, then matching distance is defined by NORM_HAMMING2.

詳情:
https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_feature2d/py_orb/py_orb.html#orb
'''


def orb_feature(image):
    image = cv2.imread(image, 0)

    ORB = cv2.ORB_create()

    kp = ORB.detect(image, None)

    kp, des = ORB.compute(image, kp)

    image = cv2.drawKeypoints(image, kp, image, color=(0, 255, 0), flags=0)

    return image


def harris(image):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    gray = np.float32(gray)
    dst = cv2.cornerHarris(gray, 2, 3, 0.04)
    # result is dilated for marking the corners, not important
    dst = cv2.dilate(dst, None)
    # Threshold for an optimal value, it may vary depending on the image.
    image[dst > 0.01 * dst.max()] = [0, 0, 255]
    return image



import cv2

'''
負責處理事件
'''


class cv_event(object):
    """
    event:
    [
    'event_FLAG_ALTKEY',
    'event_FLAG_CTRLKEY',
    'event_FLAG_LBUTTON',
    'event_FLAG_MBUTTON',
    'event_FLAG_RBUTTON',
    'event_FLAG_SHIFTKEY',
    'event_LBUTTONDBLCLK',
    'event_LBUTTONDOWN',
    'event_LBUTTONUP',
    'event_MBUTTONDBLCLK',
    'event_MBUTTONDOWN',
    'event_MBUTTONUP',
    'event_MOUSEHWHEEL',
    'event_MOUSEMOVE',
    'event_MOUSEWHEEL',
    'event_RBUTTONDBLCLK',
    'event_RBUTTONDOWN',
    'event_RBUTTONUP'
      ]

    """

    def __init__(self):
        self.ALTKEY_event = self.default_event
        self.CTRLKEY_event = self.default_event
        self.LBUTTON_event = self.default_event
        self.MBUTTON_event = self.default_event
        self.RBUTTON_event = self.default_event
        self.SHIFTKEY_event = self.default_event
        self.LBUTTONDBLCLK_event = self.default_event
        self.LBUTTONDOWN_event = self.default_event
        self.LBUTTONUP_event = self.default_event
        self.MBUTTONDBLCLK_event = self.default_event
        self.MBUTTONDOWN_event = self.default_event
        self.MBUTTONUP_event = self.default_event
        self.MOUSEHWHEEL_event = self.default_event
        self.MOUSEMOVE_event = self.default_event
        self.MOUSEWHEEL_event = self.default_event
        self.RBUTTONDBLCLK_event = self.default_event
        self.RBUTTONDOWN_event = self.default_event
        self.RBUTTONUP_event = self.default_event

    '''
    回調都會送五個參數
    def draw_circle(event,x,y,flag,param):
        if event == cv2.event_LBUTTONDBLCLK:
            cv2.circle(a.Draw.Canvas,(x,y),100,(255,0,0),-1)
    '''

    def default_event(self):
        pass

    def set_altkey_event(self, event):
        self.ALTKEY_event = event

    def altkey_event(self):
        self.ALTKEY_event()

    def Set_CTRLKEY_event(self, event):
        self.CTRLKEY_event = event

    def CTRLKEY_event(self):
        self.CTRLKEY_event()

    def Set_LBUTTON_event(self, event):
        self.LBUTTON_event = event

    def LBUTTON_event(self):
        self.LBUTTON_event()

    def Set_MBUTTON_event(self, event):
        self.MBUTTON_event = event

    def MBUTTON_event(self):
        self.MBUTTON_event()

    def Set_RBUTTON_event(self, event):
        self.RBUTTON_event = event

    def RBUTTON_event(self):
        self.RBUTTON_event()

    def Set_SHIFTKEY_event(self, event):
        self.SHIFTKEY_event = event

    def SHIFTKEY_event(self):
        self.SHIFTKEY_event()

    def Set_LBUTTONDBLCLK_event(self, event):
        self.LBUTTONDBLCLK_event = event

    def LBUTTONDBLCLK_event(self):
        self.LBUTTONDBLCLK_event()

    def Set_LBUTTONDOWN_event(self, event):
        self.LBUTTONDOWN_event = event

    def LBUTTONUP_event(self):
        self.LBUTTONUP_event()

    def Set_LBUTTONUP_event(self, event):
        self.LBUTTONUP_event = event

    def Set_MBUTTONDBLCLK_event(self, event):
        self.MBUTTONDBLCLK_event = event

    def MBUTTONDBLCLK_event(self):
        self.MBUTTONDBLCLK_event()

    def Set_MBUTTONDOWN_event(self, event):
        self.MBUTTONDOWN_event = event

    def MBUTTONDOWN_event(self):
        self.MBUTTONDOWN_event()

    def Set_MBUTTONUP_event(self, event):
        self.MBUTTONUP_event = event

    def MBUTTONUP_event(self):
        self.MBUTTONUP_event()

    def Set_MOUSEHWHEEL_event(self, event):
        self.MOUSEHWHEEL_event = event

    def MOUSEHWHEEL_event(self):
        self.MOUSEHWHEEL_event()

    def Set_MOUSEMOVE_event(self, event):
        self.MOUSEMOVE_event = event

    def MOUSEMOVE_event(self):
        self.MOUSEMOVE_event()

    def Set_MOUSEWHEEL_event(self, event):
        self.MOUSEWHEEL_event = event

    def MOUSEWHEEL_event(self):
        self.MOUSEWHEEL_event()

    def Set_RBUTTONDBLCLK_event(self, event):
        self.RBUTTONDBLCLK_event = event

    def RBUTTONDBLCLK_event(self):
        self.RBUTTONDBLCLK_event()

    def Set_RBUTTONDOWN_event(self, event):
        self.RBUTTONDOWN_event = event

    def RBUTTONDOWN_event(self):
        self.RBUTTONDOWN_event()

    def Set_RBUTTONUP_event(self, event):
        self.RBUTTONUP_event = event

    def RBUTTONUP_event(self):
        self.RBUTTONUP_event()

    def events(self):
        return [i for i in dir(cv2) if 'event' in i]

    def set_event(self, event, x, y, flags, param):

        if event == cv2.event_FLAG_ALTKEY:
            self.ALTKEY_event()

        elif event == cv2.event_FLAG_CTRLKEY:
            self.CTRLKEY_event()

        elif event == cv2.event_FLAG_LBUTTON:
            self.LBUTTON_event()

        elif event == cv2.event_FLAG_MBUTTON:
            self.MBUTTON_event()

        elif event == cv2.event_FLAG_RBUTTON:
            self.RBUTTON_event()

        elif event == cv2.event_FLAG_SHIFTKEY:
            self.SHIFTKEY_event()

        elif event == cv2.event_LBUTTONDBLCLK:
            self.LBUTTONDBLCLK_event()

        elif event == cv2.event_LBUTTONDOWN:
            self.LBUTTONDOWN_event()

        elif event == cv2.event_LBUTTONUP:
            self.LBUTTONUP_event()

        elif event == cv2.event_MBUTTONDBLCLK:
            self.MBUTTONDBLCLK_event()

        elif event == cv2.event_MBUTTONDOWN:
            self.MBUTTONDOWN_event()

        elif event == cv2.event_MBUTTONUP:
            self.MBUTTONUP_event()

        elif event == cv2.event_MOUSEHWHEEL:
            self.MOUSEHWHEEL_event()

        elif event == cv2.event_MOUSEMOVE:
            self.MOUSEMOVE_event()

        elif event == cv2.event_MOUSEWHEEL:
            self.MOUSEHWHEEL_event()

        elif event == cv2.event_RBUTTONDBLCLK:
            self.RBUTTONDBLCLK_event()

        elif event == cv2.event_RBUTTONDOWN:
            self.RBUTTONDOWN_event()

        elif event == cv2.event_RBUTTONUP:
            self.RBUTTONUP_event()

    def set_callback_event(self, Window_Name):
        cv2.setMouseCallback(Window_Name, self.set_event)

    def set_callback_function(self, Window_Name, Job_Function):
        cv2.setMouseCallback(Window_Name, Job_Function)



from je_open_cv.modules import contours
from je_open_cv.modules import cv_event
from je_open_cv.modules import draw
from je_open_cv.modules import edge
from je_open_cv.modules import face_detection
from je_open_cv.modules import feature
from je_open_cv.modules import feature_detection
from je_open_cv.modules import fourier
from je_open_cv.modules import fps
from je_open_cv.modules import gradients
from je_open_cv.modules import histograms
from je_open_cv.modules import hough_line
from je_open_cv.modules import image_operations
from je_open_cv.modules import io
from je_open_cv.modules import morphological
from je_open_cv.modules import object_tracking
from je_open_cv.modules import pyramids
from je_open_cv.modules import smoothing
from je_open_cv.modules import template_detection
from je_open_cv.modules import thresholding
from je_open_cv.modules import ui
from je_open_cv.modules import video



import cv2
import numpy as np

'''
基本的型態改變
'''

'''
The basic idea of erosion is just like soil erosion only, 
 it erodes away the boundaries of foreground object (Always try to keep foreground in white). So what does it do? 
  The kernel slides through the image (as in 2D convolution). 
   A pixel in the original image (either 1 or 0) will be considered 1 only if all the pixels under the kernel is 1,
    otherwise it is eroded (made to zero).

So what happends is that, all the pixels near boundary will be discarded depending upon the size of kernel. 
  So the thickness or size of the foreground object decreases or simply white region decreases in the image. 
   It is useful for removing small white noises (as we have seen in colorspace chapter), detach two connected objects etc.
'''


# 侵蝕
def erosion(image, kernel=(5, 5)):
    return cv2.erode(image, np.ones(kernel, np.uint8), iterations=1)


'''
It is just opposite of erosion. 
Here, a pixel element is ‘1’ if atleast one pixel under the kernel is ‘1’. 
So it increases the white region in the image or size of foreground object increases. 
Normally, in cases like noise removal, erosion is followed by dilation. 
Because, erosion removes white noises, but it also shrinks our object.
So we dilate it. Since noise is gone, they won’t come back, but our object area increases. 
It is also useful in joining broken parts of an object.
'''


# 擴張
def dilation(image, kernel=(5, 5)):
    return cv2.dilate(image, np.ones(kernel, np.uint8), iterations=1)


'''
Opening is just another name of erosion followed by dilation. 
It is useful in removing noise, as we explained above. 
Here we use the function, cv2.morphologyEx()
'''


# 侵蝕 + 擴張
def opening(image, kernel=(5, 5)):
    return cv2.morphologyEx(image, cv2.MORPH_OPEN, np.ones(kernel, np.uint8))


'''
Closing is reverse of Opening, Dilation followed by Erosion. 
It is useful in closing small holes inside the foreground objects, or small black points on the object.
'''


# 侵蝕+膨脹
def closing(image, kernel=(5, 5)):
    return cv2.morphologyEx(image, cv2.MORPH_CLOSE, np.ones(kernel, np.uint8))


'''
It is the difference between dilation and erosion of an image.
The result will look like the outline of the object.
'''


def morphological(image, kernel=(5, 5)):
    return cv2.morphologyEx(image, cv2.MORPH_GRADIENT, np.ones(kernel, np.uint8))


'''
It is the difference between input image and Opening of the image.
'''


def top_hat(image, kernel=(5, 5)):
    return cv2.morphologyEx(image, cv2.MORPH_TOPHAT, np.ones(kernel, np.uint8))


'''
It is the difference between the closing of the input image and input image.
'''


def black_hat(image, kernel=(5, 5)):
    return cv2.morphologyEx(image, cv2.MORPH_BLACKHAT, np.ones(kernel, np.uint8))


'''
We manually created a structuring elements in the previous examples with help of Numpy. 
It is rectangular shape. But in some cases, you may need elliptical/circular shaped kernels. 
So for this purpose, OpenCV has a function, cv2.getStructuringElement(). 
You just pass the shape and size of the kernel, you get the desired kernel.
'''


def get_kernel(kernel=(5, 5)):
    return cv2.getStructuringElement(cv2.MORPH_RECT, kernel)



import cv2

'''
用OpenCV做的模塊
'''

'''
switch_function 要有一個參數 會發送目前位置給Switch_Function的參數
'''


def add_trackbar(canvas_name, switch_function, track_name='name', start=0, end=255):
    cv2.createTrackbar(track_name, canvas_name, start, end, switch_function)


def get_trackbar_pos(track_name, canvas_name):
    return cv2.getTrackbarPos(track_name, canvas_name)



import cv2
import numpy as np

'''
圖像漸變用
'''

'''
OpenCV provides three types of gradient filters or High-pass filters, Sobel, Scharr and Laplacian.
'''


# 拉普拉斯算法 取XY方向
def laplacian(image):
    return cv2.Laplacian(image, cv2.CV_64F)


# SobelX 取 x方向
def sobel_x(image):
    return cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=5)


# SobelY 取 Y方向
def sobel_y(image):
    return cv2.Sobel(image, cv2.CV_64F, 0, 1, ksize=5)


'''
output datatype is cv2.CV_8U or np.uint8. But there is a slight problem with that. 
Black-to-White transition is taken as Positive slope (it has a positive value) 
while White-to-Black transition is taken as a Negative slope (It has negative value). 
So when you convert data to np.uint8, all negative slopes are made zero. 
In simple words, you miss that edge.

If you want to detect both edges, better option is to keep the output datatype to some higher forms, 
like cv2.CV_16S, cv2.CV_64F etc, take its absolute value and then convert back to cv2.CV_8U. 
Below code demonstrates this procedure for a horizontal Sobel filter and difference in results.
'''


# cv2.CV_8U
def cv_8_u(image):
    return cv2.Sobel(image, cv2.CV_8U, 1, 0, ksize=5)


# cv2.CV_64F. Then take its absolute and convert to cv2.CV_8U
def cv_64_to_cv_8_u(image):
    sobel_x64f = cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=5)
    abs_sobel64f = np.absolute(sobel_x64f)
    return np.uint8(abs_sobel64f)



import cv2

'''
邊緣檢測用
'''

"""
First argument is our input image. 
Second and third arguments are our minVal and maxVal respectively. 
Third argument is aperture_size. 
It is the size of Sobel kernel used for find image gradients. 
By default it is 3. Last argument is L2gradient which specifies the equation for finding gradient magnitude. 
If it is True, it uses the equation mentioned above which is more accurate, By default, it is False.
"""


# Canny 邊緣檢測
def Canny_Edge(Image, Min=100, Max=200):
    return cv2.Canny(Image, Min, Max)
